"""
Factory functions for technical analysis indicators.

Indicator factory names are all upper case.

Indicators offer a composable interface where a calculation routine
is bound together with its calculation parameters.

An indicator object is a callable that can be applied to prices or series data.

Indicators can be chained with the `@` operator as in `ROC(1) @ SMA(20)`.

The `@` operator can also be used to apply an indicator to its parameter.

So for example `SMA(50) @ prices` can be used to compute the 50 period simple moving average on `prices`,
instead of the more verbose `SMA(50)(prices)`.
"""

# PREAMBLE Do not edit! This file was generated

import inspect

from . import core
from .model import FuncIndicator

nan = float('nan')

def wrap_indicator(calc_func):
    """Decorator to wrap indicators"""

    def decorator(func):
        name = func.__name__
        sig = inspect.signature(func)

        def wrapper(*args, **kwargs):
            binding = sig.bind(*args, **kwargs)
            binding.apply_defaults()
            params = dict(binding.arguments)

            return FuncIndicator(
                name=name,
                func=calc_func,
                params=params,
            )

        wrapper.__name__ = func.__name__
        wrapper.__qualname__ = func.__qualname__
        wrapper.__doc__ = calc_func.__doc__
        wrapper.__signature__ = sig

        return wrapper

    return decorator


@wrap_indicator(core.calc_abs)
def ABS(*, item: str = None): ...

@wrap_indicator(core.calc_adx)
def ADX(period: int = 14): ...

@wrap_indicator(core.calc_alma)
def ALMA(period: int = 9, offset: float = 0.85, sigma: float = 6.0, *, item: str = None): ...

@wrap_indicator(core.calc_atr)
def ATR(period: int = 14): ...

@wrap_indicator(core.calc_avgprice)
def AVGPRICE(): ...

@wrap_indicator(core.calc_bbands)
def BBANDS(period: int = 20, nbdev: float = 2.0): ...

@wrap_indicator(core.calc_bbp)
def BBP(period: int = 20, nbdev: float = 2.0): ...

@wrap_indicator(core.calc_bbw)
def BBW(period: int = 20, nbdev: float = 2.0): ...

@wrap_indicator(core.calc_bop)
def BOP(period: int = 20): ...

@wrap_indicator(core.calc_cci)
def CCI(period: int = 20): ...

@wrap_indicator(core.calc_clag)
def CLAG(period: int = 1, *, item: str = None): ...

@wrap_indicator(core.calc_cmf)
def CMF(period: int = 20): ...

@wrap_indicator(core.calc_crossover)
def CROSSOVER(level: float = 0.0, *, item: str = None): ...

@wrap_indicator(core.calc_crossunder)
def CROSSUNDER(level: float = 0.0, *, item: str = None): ...

@wrap_indicator(core.calc_curve)
def CURVE(period: int = 20, *, item: str = None): ...

@wrap_indicator(core.calc_dema)
def DEMA(period: int, *, item: str = None): ...

@wrap_indicator(core.calc_diff)
def DIFF(period: int = 1, *, item: str = None): ...

@wrap_indicator(core.calc_dmi)
def DMI(period: int = 14): ...

@wrap_indicator(core.calc_ema)
def EMA(period: int, *, adjust: bool = False, item: str = None): ...

@wrap_indicator(core.calc_eval)
def EVAL(expr: str, *, as_flag: bool = False): ...

@wrap_indicator(core.calc_exp)
def EXP(*, item: str = None): ...

@wrap_indicator(core.calc_flag)
def FLAG(*, item: str = None): ...

@wrap_indicator(core.calc_hma)
def HMA(period: int, *, item: str = None): ...

@wrap_indicator(core.calc_kama)
def KAMA(period: int = 10, fastn: int = 2, slown: int = 30, *, item: str = None): ...

@wrap_indicator(core.calc_keltner)
def KELTNER(period: int = 20, nbatr: float = 2.0): ...

@wrap_indicator(core.calc_ker)
def KER(period: int = 10, *, item: str = None): ...

@wrap_indicator(core.calc_lag)
def LAG(period: int, *, item: str = None): ...

@wrap_indicator(core.calc_log)
def LOG(*, item: str = None): ...

@wrap_indicator(core.calc_lroc)
def LROC(period: int = 1, *, item: str = None): ...

@wrap_indicator(core.calc_macd)
def MACD(n1: int = 12, n2: int = 26, n3: int = 9, *, item: str = None): ...

@wrap_indicator(core.calc_macdv)
def MACDV(n1: int = 12, n2: int = 26, n3: int = 9): ...

@wrap_indicator(core.calc_mad)
def MAD(period: int = 14, *, item: str = None): ...

@wrap_indicator(core.calc_mav)
def MAV(period: int = 20, *, ma_type: str = 'SMA', item: str = None): ...

@wrap_indicator(core.calc_max)
def MAX(period: int, *, item: str = None): ...

@wrap_indicator(core.calc_mdi)
def MDI(period: int = 14): ...

@wrap_indicator(core.calc_mfi)
def MFI(period: int = 14): ...

@wrap_indicator(core.calc_midprice)
def MIDPRICE(): ...

@wrap_indicator(core.calc_min)
def MIN(period: int, *, item: str = None): ...

@wrap_indicator(core.calc_natr)
def NATR(period: int = 14): ...

@wrap_indicator(core.calc_pdi)
def PDI(period: int = 14): ...

@wrap_indicator(core.calc_ppo)
def PPO(n1: int = 12, n2: int = 26, n3: int = 9, *, item: str = None): ...

@wrap_indicator(core.calc_price)
def PRICE(item: str = None): ...

@wrap_indicator(core.calc_qsf)
def QSF(period: int = 20, offset: int = 0, *, item: str = None): ...

@wrap_indicator(core.calc_rma)
def RMA(period: int, *, item: str = None): ...

@wrap_indicator(core.calc_roc)
def ROC(period: int = 1, *, item: str = None): ...

@wrap_indicator(core.calc_rsi)
def RSI(period: int = 14, *, item: str = None): ...

@wrap_indicator(core.calc_rvalue)
def RVALUE(period: int = 20, *, item: str = None): ...

@wrap_indicator(core.calc_sar)
def SAR(afs: float = 0.02, maxaf: float = 0.2): ...

@wrap_indicator(core.calc_shift)
def SHIFT(period: int, *, item: str = None): ...

@wrap_indicator(core.calc_sign)
def SIGN(na_value: float = nan, *, item: str = None): ...

@wrap_indicator(core.calc_slope)
def SLOPE(period: int = 20, *, item: str = None): ...

@wrap_indicator(core.calc_sma)
def SMA(period: int, *, item: str = None): ...

@wrap_indicator(core.calc_stdev)
def STDEV(period: int = 20, *, item: str = None): ...

@wrap_indicator(core.calc_step)
def STEP(threshold: 'float' = 1.0, *, item: str = None): ...

@wrap_indicator(core.calc_stoch)
def STOCH(period: int = 14, fastn: int = 3, slown: int = 3): ...

@wrap_indicator(core.calc_streak)
def STREAK(*, item: str = None): ...

@wrap_indicator(core.calc_sum)
def SUM(period: int, *, item: str = None): ...

@wrap_indicator(core.calc_tema)
def TEMA(period: int = 20, *, item: str = None): ...

@wrap_indicator(core.calc_trange)
def TRANGE(*, log_prices: bool = False, percent: bool = False): ...

@wrap_indicator(core.calc_tsf)
def TSF(period: int = 20, offset: int = 0, *, item: str = None): ...

@wrap_indicator(core.calc_typprice)
def TYPPRICE(): ...

@wrap_indicator(core.calc_updown)
def UPDOWN(up_level: float = 0.0, down_level: float = 0.0, *, item: str = None): ...

@wrap_indicator(core.calc_wclprice)
def WCLPRICE(): ...

@wrap_indicator(core.calc_wma)
def WMA(period: int, *, item: str = None): ...

__all__ = [
    'ABS', 'ADX', 'ALMA', 'ATR', 'AVGPRICE', 'BBANDS', 'BBP', 'BBW', 'BOP',
    'CCI', 'CLAG', 'CMF', 'CROSSOVER', 'CROSSUNDER', 'CURVE', 'DEMA',
    'DIFF', 'DMI', 'EMA', 'EVAL', 'EXP', 'FLAG', 'HMA', 'KAMA', 'KELTNER',
    'KER', 'LAG', 'LOG', 'LROC', 'MACD', 'MACDV', 'MAD', 'MAV', 'MAX',
    'MDI', 'MFI', 'MIDPRICE', 'MIN', 'NATR', 'PDI', 'PPO', 'PRICE', 'QSF',
    'RMA', 'ROC', 'RSI', 'RVALUE', 'SAR', 'SHIFT', 'SIGN', 'SLOPE', 'SMA',
    'STDEV', 'STEP', 'STOCH', 'STREAK', 'SUM', 'TEMA', 'TRANGE', 'TSF',
    'TYPPRICE', 'UPDOWN', 'WCLPRICE', 'WMA'
]



"""
Calculation functions for technical analysis indicators.

The function names are all lower case and may conflict with standard functions,
so the best way to use this module is to alias it to a short name
like `ta` and access all functions as attributes.

The first parameter `series` or `prices` indicates whether the function
accepts a single series or a prices dataframe.

Functions that accept a series usually have an optional parameter `item`
to specify which column to use if the input is a dataframe.

All functions wrap their output to match the type of their input.

In particular the result of a function applied to a pandas series or dataframes
will have the same index as the input.
"""

# PREAMBLE Do not edit! This file was generated

import inspect

from . import core
from .core import get_series, wrap_result, column_accessor

nan = float('nan')

def wrap_function(calc_func):

    sig = inspect.signature(calc_func)
    first_param = next(iter(sig.parameters))

    def decorator(func):
        sig = inspect.signature(func)

        def wrapper(prices, *args, **kwargs):
            item = kwargs.pop('item', None)

            if first_param == 'series':
                input = get_series(prices, item)
            else:
                input = column_accessor(prices)

            result = calc_func(input, *args, **kwargs)
            return wrap_result(result, prices)

        wrapper.__name__ = func.__name__
        wrapper.__qualname__ = func.__qualname__
        wrapper.__doc__ = calc_func.__doc__
        wrapper.__signature__ = sig

        return wrapper
    return decorator


@wrap_function(core.calc_abs)
def abs(prices, *, item: str = None): ...

@wrap_function(core.calc_adx)
def adx(prices, period: int = 14): ...

@wrap_function(core.calc_alma)
def alma(prices, period: int = 9, offset: float = 0.85, sigma: float = 6.0, *, item: str = None): ...

@wrap_function(core.calc_atr)
def atr(prices, period: int = 14): ...

@wrap_function(core.calc_avgprice)
def avgprice(prices): ...

@wrap_function(core.calc_bbands)
def bbands(prices, period: int = 20, nbdev: float = 2.0): ...

@wrap_function(core.calc_bbp)
def bbp(prices, period: int = 20, nbdev: float = 2.0): ...

@wrap_function(core.calc_bbw)
def bbw(prices, period: int = 20, nbdev: float = 2.0): ...

@wrap_function(core.calc_bop)
def bop(prices, period: int = 20): ...

@wrap_function(core.calc_cci)
def cci(prices, period: int = 20): ...

@wrap_function(core.calc_clag)
def clag(prices, period: int = 1, *, item: str = None): ...

@wrap_function(core.calc_cmf)
def cmf(prices, period: int = 20): ...

@wrap_function(core.calc_crossover)
def crossover(prices, level: float = 0.0, *, item: str = None): ...

@wrap_function(core.calc_crossunder)
def crossunder(prices, level: float = 0.0, *, item: str = None): ...

@wrap_function(core.calc_curve)
def curve(prices, period: int = 20, *, item: str = None): ...

@wrap_function(core.calc_dema)
def dema(prices, period: int, *, item: str = None): ...

@wrap_function(core.calc_diff)
def diff(prices, period: int = 1, *, item: str = None): ...

@wrap_function(core.calc_dmi)
def dmi(prices, period: int = 14): ...

@wrap_function(core.calc_ema)
def ema(prices, period: int, *, adjust: bool = False, item: str = None): ...

@wrap_function(core.calc_eval)
def eval(prices, expr: str, *, as_flag: bool = False): ...

@wrap_function(core.calc_exp)
def exp(prices, *, item: str = None): ...

@wrap_function(core.calc_flag)
def flag(prices, *, item: str = None): ...

@wrap_function(core.calc_hma)
def hma(prices, period: int, *, item: str = None): ...

@wrap_function(core.calc_kama)
def kama(prices, period: int = 10, fastn: int = 2, slown: int = 30, *, item: str = None): ...

@wrap_function(core.calc_keltner)
def keltner(prices, period: int = 20, nbatr: float = 2.0): ...

@wrap_function(core.calc_ker)
def ker(prices, period: int = 10, *, item: str = None): ...

@wrap_function(core.calc_lag)
def lag(prices, period: int, *, item: str = None): ...

@wrap_function(core.calc_log)
def log(prices, *, item: str = None): ...

@wrap_function(core.calc_lroc)
def lroc(prices, period: int = 1, *, item: str = None): ...

@wrap_function(core.calc_macd)
def macd(prices, n1: int = 12, n2: int = 26, n3: int = 9, *, item: str = None): ...

@wrap_function(core.calc_macdv)
def macdv(prices, n1: int = 12, n2: int = 26, n3: int = 9): ...

@wrap_function(core.calc_mad)
def mad(prices, period: int = 14, *, item: str = None): ...

@wrap_function(core.calc_mav)
def mav(prices, period: int = 20, *, ma_type: str = 'SMA', item: str = None): ...

@wrap_function(core.calc_max)
def max(prices, period: int, *, item: str = None): ...

@wrap_function(core.calc_mdi)
def mdi(prices, period: int = 14): ...

@wrap_function(core.calc_mfi)
def mfi(prices, period: int = 14): ...

@wrap_function(core.calc_midprice)
def midprice(prices): ...

@wrap_function(core.calc_min)
def min(prices, period: int, *, item: str = None): ...

@wrap_function(core.calc_natr)
def natr(prices, period: int = 14): ...

@wrap_function(core.calc_pdi)
def pdi(prices, period: int = 14): ...

@wrap_function(core.calc_ppo)
def ppo(prices, n1: int = 12, n2: int = 26, n3: int = 9, *, item: str = None): ...

@wrap_function(core.calc_price)
def price(prices, item: str = None): ...

@wrap_function(core.calc_qsf)
def qsf(prices, period: int = 20, offset: int = 0, *, item: str = None): ...

@wrap_function(core.calc_rma)
def rma(prices, period: int, *, item: str = None): ...

@wrap_function(core.calc_roc)
def roc(prices, period: int = 1, *, item: str = None): ...

@wrap_function(core.calc_rsi)
def rsi(prices, period: int = 14, *, item: str = None): ...

@wrap_function(core.calc_rvalue)
def rvalue(prices, period: int = 20, *, item: str = None): ...

@wrap_function(core.calc_sar)
def sar(prices, afs: float = 0.02, maxaf: float = 0.2): ...

@wrap_function(core.calc_shift)
def shift(prices, period: int, *, item: str = None): ...

@wrap_function(core.calc_sign)
def sign(prices, na_value: float = nan, *, item: str = None): ...

@wrap_function(core.calc_slope)
def slope(prices, period: int = 20, *, item: str = None): ...

@wrap_function(core.calc_sma)
def sma(prices, period: int, *, item: str = None): ...

@wrap_function(core.calc_stdev)
def stdev(prices, period: int = 20, *, item: str = None): ...

@wrap_function(core.calc_step)
def step(prices, threshold: 'float' = 1.0, *, item: str = None): ...

@wrap_function(core.calc_stoch)
def stoch(prices, period: int = 14, fastn: int = 3, slown: int = 3): ...

@wrap_function(core.calc_streak)
def streak(prices, *, item: str = None): ...

@wrap_function(core.calc_sum)
def sum(prices, period: int, *, item: str = None): ...

@wrap_function(core.calc_tema)
def tema(prices, period: int = 20, *, item: str = None): ...

@wrap_function(core.calc_trange)
def trange(prices, *, log_prices: bool = False, percent: bool = False): ...

@wrap_function(core.calc_tsf)
def tsf(prices, period: int = 20, offset: int = 0, *, item: str = None): ...

@wrap_function(core.calc_typprice)
def typprice(prices): ...

@wrap_function(core.calc_updown)
def updown(prices, up_level: float = 0.0, down_level: float = 0.0, *, item: str = None): ...

@wrap_function(core.calc_wclprice)
def wclprice(prices): ...

@wrap_function(core.calc_wma)
def wma(prices, period: int, *, item: str = None): ...

__all__ = [
    'abs', 'adx', 'alma', 'atr', 'avgprice', 'bbands', 'bbp', 'bbw', 'bop',
    'cci', 'clag', 'cmf', 'crossover', 'crossunder', 'curve', 'dema',
    'diff', 'dmi', 'ema', 'eval', 'exp', 'flag', 'hma', 'kama', 'keltner',
    'ker', 'lag', 'log', 'lroc', 'macd', 'macdv', 'mad', 'mav', 'max',
    'mdi', 'mfi', 'midprice', 'min', 'natr', 'pdi', 'ppo', 'price', 'qsf',
    'rma', 'roc', 'rsi', 'rvalue', 'sar', 'shift', 'sign', 'slope', 'sma',
    'stdev', 'step', 'stoch', 'streak', 'sum', 'tema', 'trange', 'tsf',
    'typprice', 'updown', 'wclprice', 'wma'
]



"""
The `mintalib` package contains four main modules:

- `mintalib.core`:
    core calculation rountines implemented in cython, with names like `calc_sma`, `calc_ema`, etc ...  
- `mintalib.functions`:
    wrapper functions to compute calculations on series and dataframes, with names like `sma`, `ema`, etc ...
- `mintalib.indicators`
    composable interface to indicators with names like `SMA`, `EMA`, etc ...
- `mintalib.expressions`
    polars expressions library with names like `SMA`, `EMA`, etc ...
"""




"""Expressions Module"""

# PREAMBLE Do not edit! This file was generated

import inspect

import polars as pl

from polars.datatypes import Struct, Float64

from . import core

from typing import Union, TypeAlias


IntoExpr: TypeAlias = Union[pl.Expr, str, pl.Series]

nan = float('nan')


def get_series_expr(src):
    if isinstance(src, str):
        return pl.col(src)
    elif isinstance(src, pl.Expr):
        return src
    else:
        raise ValueError("src must be a string or a Polars Series.")


def get_struct_expr(src):
    if src is None:
        return pl.struct("*")
    
    if isinstance(src, str):
        return pl.col(src).struct
    
    if isinstance(src, pl.Expr):
        return pl.struct(src)
        
    raise ValueError(f"Unsupported src type: {type(src)}")


def wrap_expression(calc_func):
    calc_sig = inspect.signature(calc_func)
    first_param = next(iter(calc_sig.parameters.values()))
    force_struct = first_param.name == 'prices'

    def decorator(func):
        name = func.__name__.lower()
        metadata = getattr(calc_func, 'metadata', {})
        output_names = metadata.get('output_names', ())
        output_type = Struct({n: Float64 for n in output_names}) if output_names else Float64
        signature = inspect.signature(func)

        def wrapper(*args, **kwargs):
            bound_args = signature.bind(*args, **kwargs)
            args, kwargs = (), dict(bound_args.arguments)

            src = kwargs.pop('src', None)

            if force_struct:
                source = get_struct_expr(src)
            else:
                source = get_series_expr(src)

            def batch_func(prices):
                if force_struct:
                    prices = prices.struct.unnest()

                output = calc_func(prices, *args, **kwargs)
                
                if isinstance(output, tuple):
                    return pl.DataFrame(output._asdict()).fill_nan(None).to_struct()
                else:
                    return pl.Series(output).fill_nan(None)
            
            expr = source.map_batches(batch_func, return_dtype=output_type).alias(name)
#            expr = expr.struct.unnest() if output_names else expr.alias(name)
            
            return expr
        
        wrapper.__name__ = func.__name__
        wrapper.__qualname__ = func.__qualname__
        wrapper.__doc__ = calc_func.__doc__
        wrapper.__signature__ = signature

        return wrapper
    
    return decorator

@wrap_expression(core.calc_abs)
def ABS(src): ...


@wrap_expression(core.calc_adx)
def ADX(period: int = 14, *, src='*'): ...


@wrap_expression(core.calc_alma)
def ALMA(src, period: int = 9, offset: float = 0.85, sigma: float = 6.0): ...


@wrap_expression(core.calc_atr)
def ATR(period: int = 14, *, src='*'): ...


@wrap_expression(core.calc_avgprice)
def AVGPRICE(*, src='*'): ...


@wrap_expression(core.calc_bbands)
def BBANDS(period: int = 20, nbdev: float = 2.0, *, src='*'): ...


@wrap_expression(core.calc_bop)
def BOP(period: int = 20, *, src='*'): ...


@wrap_expression(core.calc_cci)
def CCI(period: int = 20, *, src='*'): ...


@wrap_expression(core.calc_clag)
def CLAG(src, period: int = 1): ...


@wrap_expression(core.calc_cmf)
def CMF(period: int = 20, *, src='*'): ...


@wrap_expression(core.calc_crossover)
def CROSSOVER(src, level: float = 0.0): ...


@wrap_expression(core.calc_crossunder)
def CROSSUNDER(src, level: float = 0.0): ...


@wrap_expression(core.calc_curve)
def CURVE(src, period: int = 20): ...


@wrap_expression(core.calc_dema)
def DEMA(src, period: int): ...


@wrap_expression(core.calc_diff)
def DIFF(src, period: int = 1): ...


@wrap_expression(core.calc_dmi)
def DMI(period: int = 14, *, src='*'): ...


@wrap_expression(core.calc_ema)
def EMA(src, period: int, *, adjust: bool = False): ...


@wrap_expression(core.calc_exp)
def EXP(src): ...


@wrap_expression(core.calc_flag)
def FLAG(src): ...


@wrap_expression(core.calc_hma)
def HMA(src, period: int): ...


@wrap_expression(core.calc_kama)
def KAMA(src, period: int = 10, fastn: int = 2, slown: int = 30): ...


@wrap_expression(core.calc_keltner)
def KELTNER(period: int = 20, nbatr: float = 2.0, *, src='*'): ...


@wrap_expression(core.calc_ker)
def KER(src, period: int = 10): ...


@wrap_expression(core.calc_lag)
def LAG(src, period: int): ...


@wrap_expression(core.calc_log)
def LOG(src): ...


@wrap_expression(core.calc_lroc)
def LROC(src, period: int = 1): ...


@wrap_expression(core.calc_macd)
def MACD(src, n1: int = 12, n2: int = 26, n3: int = 9): ...


@wrap_expression(core.calc_mad)
def MAD(src, period: int = 14): ...


@wrap_expression(core.calc_mav)
def MAV(src, period: int = 20, *, ma_type: str = 'SMA'): ...


@wrap_expression(core.calc_max)
def MAX(src, period: int): ...


@wrap_expression(core.calc_mdi)
def MDI(period: int = 14, *, src='*'): ...


@wrap_expression(core.calc_mfi)
def MFI(period: int = 14, *, src='*'): ...


@wrap_expression(core.calc_midprice)
def MIDPRICE(*, src='*'): ...


@wrap_expression(core.calc_min)
def MIN(src, period: int): ...


@wrap_expression(core.calc_natr)
def NATR(period: int = 14, *, src='*'): ...


@wrap_expression(core.calc_pdi)
def PDI(period: int = 14, *, src='*'): ...


@wrap_expression(core.calc_ppo)
def PPO(src, n1: int = 12, n2: int = 26, n3: int = 9): ...


@wrap_expression(core.calc_price)
def PRICE(item: str = None, *, src='*'): ...


@wrap_expression(core.calc_qsf)
def QSF(src, period: int = 20, offset: int = 0): ...


@wrap_expression(core.calc_rma)
def RMA(src, period: int): ...


@wrap_expression(core.calc_roc)
def ROC(src, period: int = 1): ...


@wrap_expression(core.calc_rsi)
def RSI(src, period: int = 14): ...


@wrap_expression(core.calc_rvalue)
def RVALUE(src, period: int = 20): ...


@wrap_expression(core.calc_sar)
def SAR(afs: float = 0.02, maxaf: float = 0.2, *, src='*'): ...


@wrap_expression(core.calc_shift)
def SHIFT(src, period: int): ...


@wrap_expression(core.calc_sign)
def SIGN(src, na_value: float = nan): ...


@wrap_expression(core.calc_slope)
def SLOPE(src, period: int = 20): ...


@wrap_expression(core.calc_sma)
def SMA(src, period: int): ...


@wrap_expression(core.calc_stdev)
def STDEV(src, period: int = 20): ...


@wrap_expression(core.calc_step)
def STEP(src, threshold: 'float' = 1.0): ...


@wrap_expression(core.calc_stoch)
def STOCH(period: int = 14, fastn: int = 3, slown: int = 3, *, src='*'): ...


@wrap_expression(core.calc_streak)
def STREAK(src): ...


@wrap_expression(core.calc_sum)
def SUM(src, period: int): ...


@wrap_expression(core.calc_tema)
def TEMA(src, period: int = 20): ...


@wrap_expression(core.calc_trange)
def TRANGE(*, log_prices: bool = False, percent: bool = False, src='*'): ...


@wrap_expression(core.calc_tsf)
def TSF(src, period: int = 20, offset: int = 0): ...


@wrap_expression(core.calc_typprice)
def TYPPRICE(*, src='*'): ...


@wrap_expression(core.calc_updown)
def UPDOWN(src, up_level: float = 0.0, down_level: float = 0.0): ...


@wrap_expression(core.calc_wclprice)
def WCLPRICE(*, src='*'): ...


@wrap_expression(core.calc_wma)
def WMA(src, period: int): ...


__all__ = [
    'ABS', 'ADX', 'ALMA', 'ATR', 'AVGPRICE', 'BBANDS', 'BOP', 'CCI',
    'CLAG', 'CMF', 'CROSSOVER', 'CROSSUNDER', 'CURVE', 'DEMA', 'DIFF',
    'DMI', 'EMA', 'EXP', 'FLAG', 'HMA', 'KAMA', 'KELTNER', 'KER', 'LAG',
    'LOG', 'LROC', 'MACD', 'MAD', 'MAV', 'MAX', 'MDI', 'MFI', 'MIDPRICE',
    'MIN', 'NATR', 'PDI', 'PPO', 'PRICE', 'QSF', 'RMA', 'ROC', 'RSI',
    'RVALUE', 'SAR', 'SHIFT', 'SIGN', 'SLOPE', 'SMA', 'STDEV', 'STEP',
    'STOCH', 'STREAK', 'SUM', 'TEMA', 'TRANGE', 'TSF', 'TYPPRICE',
    'UPDOWN', 'WCLPRICE', 'WMA'
]



import pytest

from mintalib import core, functions, indicators
from mintalib.samples import sample_prices
from mintalib.testing import first_param, sample_params

try:
    import polars
    from mintalib import expressions
except ImportError:
    polars = expressions = None


def core_functions():
    return [
        k for k, v in vars(core).items()
        if k.startswith(("calc_", "flag_"))
        and callable(v)
        and first_param(v) in ("prices", "series")
    ]

def list_functions():
    return [k for k, v in vars(functions).items() if callable(v) and first_param(v) in ("prices", "series")]

def list_indicators():
    return [k for k, v in vars(indicators).items() if k.isupper() and callable(v)]

def list_expressions():
    if expressions is not None:
        return [k for k, v in vars(expressions).items() if k.isupper() and callable(v)]
    return ()



def test_samples():
    prices = sample_prices()
    assert prices is not None


@pytest.mark.parametrize("name", core_functions())
def test_core(name):
    func = getattr(core, name)

    ftype = first_param(func)
    assert ftype in ("series", "prices")

    kwds = sample_params(func)

    item = "close" if ftype == "series" else None
    data = sample_prices(item=item)

    result = func(data, **kwds)
    assert result is not None


@pytest.mark.parametrize("name", list_functions())
def test_function(name):
    func = getattr(functions, name)

    ftype = first_param(func)
    assert ftype in ("series", "prices")

    kwds = sample_params(func)

    item = "close" if ftype == "series" else None
    data = sample_prices(item=item)

    result = func(data, **kwds)
    assert result is not None


@pytest.mark.parametrize("name", list_indicators())
def test_indicator(name):
    func = getattr(indicators, name)
    kwds = sample_params(func)
    indicator = func(**kwds)

    assert callable(func)

    prices = sample_prices()

    result = indicator(prices)
    assert result is not None


@pytest.mark.skipif(polars is None, reason="requires polars")
@pytest.mark.parametrize("name", list_expressions())
def test_expression(name):
    func = getattr(expressions, name)
    kwds = sample_params(func)
    expr = func(**kwds)

    prices = sample_prices()
    prices = polars.from_pandas(prices, include_index=True)

    result = prices.select(expr)

    assert result is not None
